class SequentialFitnessCaller:
    """
    Fitness caller used for sequential implementation of NMMSO algorithm.
    """
    def __init__(self):
        self.problem = None
        self.data = []

    def set_problem(self, problem):
        """
        Sets the problem object to use to calculate the fitness.

        Arguments
        ---------

        problem
            Problem object implementing the fitness method.
        """
        self.problem = problem

    def add(self, location, userdata):
        """
        Add a location to be evaluated.

        Arguments
        ---------

        location : numpy array
            Location to be evaluated.

        userdata
            Userdata to be returned with the evaluation result.
        """
        self.data.append((location, userdata))

    def evaluate(self):
        """
        Evaluates all the locations.

        Returns
        -------

        list of (location, value, userdate) tuples
            Tuples containing the location, value and corresponding user data
        """
        result = []
        for location, userdata in self.data:
            value = self.problem.fitness(location)
            result.append((location, value, userdata))

        self.data = []
        return result

    def finish(self):
        """
        Terminates the fitness caller.
        """
        pass



class BaseListener:
    """
    Empty base class for listeners to NMMSO
    """

    def __init__(self):
        pass

    def set_nmmso(self, nmmso):
        """
        Sets the Nmmso object.

        Arguments
        ---------

        nmmso : Nmmso
            The Nmmso object.
        """
        pass

    def iteration_started(self):
        """
        Called when a new iteration is started.
        """
        pass

    def location_evaluated(self, location, value):
        """
        Called whenever a new location in parameter space has been evaluated.

        Arguments
        ---------

        location : numpy array
            Array of parameter values
        value : float
            Fitness value for this location.
        """
        pass

    def swarm_peak_changed(self, swarm, old_location, old_value):
        """
        Called when a swarm's peak location has changed.

        Arguments
        ---------

        swarm : Swarm
            The swarm that has changed.  This will contain the new location and fitness values.
        old_location : numpy array
            Previous location of the swarm.
        old_value :
            Previous fitness value of the swarm.
        """
        pass

    def swarm_created_at_random(self, new_swarm):
        """
         Called when a new swarm has been created at a random location.

        Arguments
        ---------

        new_swarm : Swarm
            The swarm that has been created.
        """
        pass

    def swarm_created_from_crossover(self, new_swarm, parent_swarm1, parent_swarm2):
        """
        Called when a new swarm has been created at a location produced by crossover from two
        other swarms.

        Arguments
        ---------

        new_swarm : Swarm
            Newly created swarm
        parent_swarm1 : Swarm
            Parent swarm
        parent_swarm2 : Swarm
            Parent swarm
        """
        pass

    def merging_started(self):
        """
        Called when the merging stage has started.
        """
        pass

    def merged_close_swarms(self, swarm1, swarm2):
        """
        Called when two swarms have been merged because them were close.

        Arguments
        ---------

        swarm1 : Swarm
            The swarm that will be kept.
        swarm2 : Swarm
            The swarm that will be merged into the other swarm.
        """
        pass

    def merged_saddle_swarms(self, swarm1, swarm2):
        """
        Called when two swarms have been merged because the midpoint between them is fitter
        than both swarms.

        Arguments
        ---------

        swarm1 : Swarm
            The swarm that will be kept.
        swarm2 : Swarm
            The swarm that will be merged into the other swarm.
        """
        pass

    def merging_ended(self):
        """
        Called when the merging stage has completed.
        """
        pass

    def incrementing_swarms_started(self):
        """
        Called when the swarm incrementing stage has started.
        """
        pass

    def swarm_added_particle(self, swarm):
        """
        Called when a particle has been added to a swarm.

        Arguments
        ---------

        swarm : Swarm
            The swarm to which the particle has been added.
        """
        pass

    def swarm_moved_particle(self, swarm):
        """
        Called when a particle has been moved.

        Arguments
        ---------

        swarm : Swarm
            The swarm in which the particle has been moved.
        """
        pass

    def incrementing_swarms_ended(self):
        """
        Called when the swarm incrementing stage has finished.
        """
        pass

    def hiving_swams_started(self):
        """
        Called when the swarm hiving stage has completed.
        """
        pass

    def hiving_new_swarm(self, new_swarm, parent_swarm):
        """
        Called when a new swarm is hived from a parent swarm.

        Arguments
        ---------

        new_swarm : Swarm
            The swarm in which been hived off.
        parent_swarm : Swarm
            The parent swarm.
        """
        pass

    def hiving_swarms_ended(self):
        """
        Called when the swarm hiving stage has completed.
        """
        pass

    def iteration_ended(
            self, n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples):
        """
        Called when an iteration has completed.

        Arguments
        ---------

        n_new_locations : int
            Number of fitness evaluations when incrementing the swarms in this iteration
        n_mid_evals : int
            Number of midpoint fitness evaluations when merging in this iteration
        n_evol_modes : int
            Number of fitness evaluations when evolving new swarms in this iteration
        n_rand_modes : int
            Number of fitness evaluations when generating a new swarm at a random location in this iteration
        n_hive_samples : int
            Number of fitness evaluation during the hive process in this iteration

        """
        pass

    def max_evaluations_reached(self):
        """
        Called when the specified maximum number of evaluation has been reached.
        """
        pass



import math
from pynmmso.listeners.base_listener import BaseListener


class ParallelPredictorListener(BaseListener):
    """
    Listener class used to predict how a parallel implementation may peform.
    """

    def __init__(self):
        self.num_workers = [2, 4, 6, 8, 10, 12, 16, 32, 64, 128]
        self.min_evaluations = [0] * len(self.num_workers)
        self.max_active_workers = 0
        self.total_evaluations = 0
        super().__init__()

    def iteration_ended(
            self, n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples):
        self.total_evaluations += \
            n_new_locations + n_mid_evals + n_evol_modes + n_rand_modes + n_hive_samples

        for i in range(len(self.num_workers)):
            self.min_evaluations[i] += n_evol_modes + n_rand_modes + n_hive_samples
            self.min_evaluations[i] += math.ceil(n_new_locations/self.num_workers[i])
            self.min_evaluations[i] += math.ceil(n_mid_evals/self.num_workers[i])
            if n_new_locations > self.max_active_workers:
                self.max_active_workers = n_new_locations
            if n_mid_evals > self.max_active_workers:
                self.max_active_workers = n_mid_evals

    def print_summary(self):
        """
        Prints a summary of the results.
        """
        print("Total number of fitness evaluations: {}".format(self.total_evaluations))
        print("Maximum number of workers: {}".format(self.max_active_workers))
        for i in range(len(self.num_workers)):
            print("{} workers, effective evaluations {}, "
                  "minimum runtime as percentage of sequential runtime: {:0.2f}%".format(
                      self.num_workers[i], self.min_evaluations[i],
                      self.min_evaluations[i]/self.total_evaluations * 100.0))



import random
import math
import numpy as np
import pynmmso.swarm as s
from pynmmso.listeners import MultiListener
from pynmmso.sequential_fitness_caller import SequentialFitnessCaller


class ModeResult:
    """
    Contains a mode result.

    Parameters
    ----------

    location: Numpy array
        Location of this mode in parameter space.

    value: float
        Fitness score of this mode.

    Attributes
    ----------

    location : Numpy array
        Location of this mode in parameter space.

    value : float
        Fitness score of this mode.

    """
    def __init__(self, location, value):
        self.location = location
        self.value = value


class Nmmso:
    """
    Niching Migratory Multi-Swarm Optimser.

    Parameters
    ----------

    problem :
        Instance of the problem class. Must implement get_bounds and fitness functions.

    swarm_size : int
        The maximum number of particles in a swarm. The default is -1 which specifies
        the number of particles will be 4 + floor(3*log(D)), where D is the number of
        dimensions of the problem.

    max_evol : int
        Maximum number of swarms that are updated in each iteration. The default is
        100.

    tol_val: float
        Tolerance of Euclidean distance between swarms. Swarms whose distance is below
        this threshold will be merged. Default value is  1e-06.

    fitness_caller :
        Used to specify various approaches to parallelism. Default is to use no parallelism.

    Attributes
    ----------

    evaluations: int
        Number of evaluations of the fitness function that have been performed so far.

    swarms:  Set of Swarm objects
        The currently active swarms.

    """
    def __init__(self, problem, swarm_size=-1, max_evol=100, tol_val=1e-06,
                 fitness_caller=SequentialFitnessCaller()):

        self.problem = problem
        self.max_evol = max_evol if max_evol > 0 else 100

        self.tol_val = tol_val
        self.min, self.max = problem.get_bounds()
        self.min = np.array(self.min)  # convert to numpy array
        self.max = np.array(self.max)  # convert to numpy array

        # Validate the input bounds
        assert np.all(len(self.min) == len(self.max)), \
            "Lower bounds list must be the same length as the upper bounds list."
        assert np.all(self.min < self.max), \
            "Problem lower bound must be less than upper bound for all dimensions."

        self.num_dimensions = len(self.min)

        if swarm_size < 1:
            self.swarm_size = 4 + math.floor(3 * math.log(self.num_dimensions))
        else:
            self.swarm_size = swarm_size

        self.converged_modes = 0
        self.next_swarm_id = 1
        self.evaluations = 0

        self.swarms = set()
        self.total_mid_evals = 0
        self.total_new_locations = 0
        self.total_evol_modes = 0
        self.total_rand_modes = 0
        self.total_hive_samples = 0

        self.fitness_caller = fitness_caller
        self.fitness_caller.set_problem(self.problem)

        self.listener = None

    def add_listener(self, listener):
        """
        Add a listener object that will receive notifications of events that occurs
        during the optimisation process.

        Parameters
        ----------

        listener : subclass of nmmso.listeners.BaseListener
            Listener object to receive notification of events.

        """
        if self.listener is None:
            self.listener = MultiListener()

        self.listener.add_listener(listener)
        listener.set_nmmso(self)

    def run(self, max_evaluations):
        """
        Runs is the optimisation algorithm until the specified number of fitness
        evaluations has been exceeded.

        If the algorithm has already been run it will continue from where it
        left off.

        Parameters
        ----------

        max_evaluations : int
            The maximum number of fitness evaluations.  The aglorithm may execute
            more evaluations that this number but will stop at the end of the
            iteration when this limit is exceeded.

        Returns
        -------

        list of ModeResult
            All the modes found by the algorithm.

        """
        while self.evaluations < max_evaluations:
            self.iterate()

        if self.listener is not None:
            self.listener.max_evaluations_reached()

        return self.get_result()

    def get_result(self):
        """
        Obtains the modes found by the algorithm.

        Returns
        -------

        list of ModeResult
            All the modes found by the algorithm.
        """
        has_remap_parameters_function = False
        remap_func = getattr(self.problem, "remap_parameters", None)
        if callable(remap_func):
            has_remap_parameters_function = True

        result = []
        for swarm in self.swarms:
            loc = swarm.mode_location
            if has_remap_parameters_function:
                loc = self.problem.remap_parameters(loc)
            result.append(ModeResult(loc, swarm.mode_value))
        return result

    def iterate(self):
        """
        Executes a single iteration of the algorithm.

        """
        if self.listener is not None:
            self.listener.iteration_started()

        if self.evaluations == 0:
            swarm = self._new_swarm()
            swarm.set_initial_location()
            swarm.evaluate_first()
            self._add_swarm(swarm)
            if self.listener is not None:
                self.listener.swarm_created_at_random(swarm)
            self.evaluations = 0
            num_of_evol_modes = 0
            num_rand_modes = 1
        else:
            # create speculative new swarm, either at random in design space, or via crossover
            if random.random() < 0.5 or \
                    len(self.swarms) == 1 or len(self.max) == 1 or self.max_evol == 1:
                num_of_evol_modes = 0
                num_rand_modes = self._random_new()  # this currently *always* returns 1
            else:
                num_rand_modes = 0
                num_of_evol_modes = self._evolve()

        if self.listener is not None:
            self.listener.merging_started()

        # See if modes should be merged together
        num_of_mid_evals = 0
        while sum([mode.changed for mode in self.swarms]) > 0:
            merge_evals = self._merge_swarms()
            num_of_mid_evals = num_of_mid_evals + merge_evals

        if self.listener is not None:
            self.listener.merging_ended()

        # Now increment the swarms
        if self.listener is not None:
            self.listener.incrementing_swarms_started()

        # if we have more than max_evol, then only increment a subset
        limit = min(self.max_evol, len(self.swarms))
        if len(self.swarms) > self.max_evol:
            if np.random.random() < 0.5:
                # select the fittest
                swarms_to_increment = \
                    sorted(self.swarms, key=lambda x: x.mode_value, reverse=True)[0:limit]
            else:
                # select at random
                swarms_to_increment = random.sample(self.swarms, limit)

        else:
            # use all swarms when incrementing
            swarms_to_increment = self.swarms

        for swarm in swarms_to_increment:
            swarm.increment()

        # evaluate new member/new location of swarm member
        num_of_new_locations = self._evaluate_new_locations(swarms_to_increment)

        if self.listener is not None:
            self.listener.incrementing_swarms_ended()

        # attempt to split off a member from one of the swarms to seed a new
        # swarm (if detected to be on another peak)
        if self.listener is not None:
            self.listener.hiving_swams_started()
        num_of_hive_samples = self._hive()
        if self.listener is not None:
            self.listener.hiving_swarms_ended()

        # update the total number of function evaluations used, with those
        # required at each of the algorithm stages
        self.evaluations = self.evaluations + \
                           num_of_mid_evals + \
                           num_of_new_locations + \
                           num_of_evol_modes + \
                           num_rand_modes + \
                           num_of_hive_samples

        self.total_mid_evals += num_of_mid_evals
        self.total_new_locations += num_of_new_locations
        self.total_evol_modes += num_of_evol_modes
        self.total_rand_modes += num_rand_modes
        self.total_hive_samples += num_of_hive_samples

        if self.listener is not None:
            self.listener.iteration_ended(
                num_of_new_locations, num_of_mid_evals, num_of_evol_modes,
                num_rand_modes, num_of_hive_samples)

    def _merge_swarms(self):

        # Only concern ourselves with modes that have actually shifted, or are new
        # since the last generation, as no need to check others
        swarms_changed = list(filter(lambda x: x.changed, self.swarms))
        self._reset_changed_flags()

        n = len(swarms_changed)

        closest_swarms = set()

        number_of_mid_evals = 0

        # only compare if there is a changed mode, and more than one mode in system
        if n >= 1 and len(self.swarms) > 1:
            for swarm in swarms_changed:
                closest_swarm, _ = swarm.find_nearest(self.swarms)
                tuple_of_swarms = Nmmso._create_swarm_tuple(swarm, closest_swarm)
                closest_swarms.add(tuple_of_swarms)
                if swarm.number_of_particles == 1:
                    swarm.initialise_new_swarm_velocities()

            to_merge = set()
            number_of_mid_evals = 0

            for swarm1, swarm2 in closest_swarms:
                if swarm1.distance_to(swarm2) < self.tol_val:
                    # merge if sufficiently close
                    to_merge.add((swarm1, swarm2, True))
                else:
                    # otherwise merge if midpoint is fitter
                    mid_loc = 0.5 * (swarm1.mode_location - swarm2.mode_location) + \
                              swarm2.mode_location
                    self.fitness_caller.add(mid_loc, (swarm1, swarm2))
                    number_of_mid_evals += 1

            # Get the results from the fitness caller
            for loc, y, userdata in self.fitness_caller.evaluate():
                swarm1, swarm2 = userdata
                if self.listener is not None:
                    self.listener.location_evaluated(loc, y)

                if y > swarm2.mode_value:
                    swarm2.update_location_and_value(loc, y)
                    to_merge.add((swarm1, swarm2, False))
                elif swarm1.mode_value <= y:
                    to_merge.add((swarm1, swarm2, False))

            # Merge the swarms - it is possible that we get one swarm involved in two merges.
            # In such situation depending on order we can get three combos:
            #   2 -> 1 and 3 -> 1  : do two merges into swarm 1, swarms 2 and 3 are deleted
            #   2 -> 1 and 3 -> 2  : merge 2 into 1, swarms 2 and 3 are deleted (3's
            #                        particles go to 2 but it is pointless)
            #   3 -> 2 and 2 -> 1  : all good, 2 and 3 are deleted but 3's particles have a
            #                        chance of reaching swarm 1
            for swarm1, swarm2, are_close in to_merge:
                # print('Merging {} and {}'.format(swarm1.id, swarm2.id))
                if swarm1.mode_value > swarm2.mode_value:
                    swarm1.merge(swarm2)
                    swarm1.changed = True
                    if swarm2 in self.swarms:
                        self.swarms.remove(swarm2)
                    if self.listener is not None:
                        if are_close:
                            self.listener.merged_close_swarms(swarm1, swarm2)
                        else:
                            self.listener.merged_saddle_swarms(swarm1, swarm2)
                else:
                    swarm2.merge(swarm1)
                    swarm2.changed = True
                    if swarm1 in self.swarms:
                        self.swarms.remove(swarm1)
                    if self.listener is not None:
                        if are_close:
                            self.listener.merged_close_swarms(swarm2, swarm1)
                        else:
                            self.listener.merged_saddle_swarms(swarm2, swarm1)

        # If only one mode, choose arbitrary dist for it
        if len(self.swarms) == 1:
            # for swarm in self.swarms:
            [swarm] = self.swarms
            swarm.set_arbitrary_distance()

        return number_of_mid_evals

    @staticmethod
    def _distance_to(location1, location2):
        """Euclidean distance between two locations"""
        return np.linalg.norm(location1-location2)

    @staticmethod
    def _create_swarm_tuple(swarm1, swarm2):
        """Create swarm tuples in a consistent way so we don't have (a,b) and
        (b,a) as separate keys"""
        if swarm1.id < swarm2.id:
            tuple_of_swarms = (swarm1, swarm2)
        else:
            tuple_of_swarms = (swarm2, swarm1)
        return tuple_of_swarms

    def _reset_changed_flags(self):
        for mode in self.swarms:
            mode.changed = False

    def _evaluate_new_locations(self, swarm_subset):
        """Evaluates the new locations of all swarms in the given subset of swarms."""

        # clear changed flag of all swarms (not just this subset)
        for swarm in self.swarms:
            swarm.changed = False

        # evaluate the subset of swarms
        for swarm in swarm_subset:
            self.fitness_caller.add(swarm.new_location, swarm)

        for loc, y, swarm in self.fitness_caller.evaluate():
            if self.listener is not None:
                self.listener.location_evaluated(loc, y)
            swarm.evaluate(y)

        return len(swarm_subset)

    def _evolve(self):
        """Evolves a new swarm by crossing over two existing swarms."""
        candidate_swarms = self.swarms

        if len(candidate_swarms) > self.max_evol and random.random() < 0.5:
            # if have a lot of swarms occasionally reduce the candidate set to be the fittest ones
            candidate_swarms = \
                sorted(candidate_swarms, key=lambda x: x.mode_value, reverse=True)[:self.max_evol]

        # select two at random from the candidate swarms
        swarms_to_cross = random.sample(candidate_swarms, 2)

        swarm = self._new_swarm()
        swarm.initialise_with_uniform_crossover(swarms_to_cross[0], swarms_to_cross[1])
        self._add_swarm(swarm)

        if self.listener is not None:
            self.listener.swarm_created_from_crossover(
                swarm, swarms_to_cross[0], swarms_to_cross[1])

        number_of_new_modes = 1
        return number_of_new_modes

    def _hive(self):

        # Comments by CW:
        #  - some of this probably could be put into the swarm class, but we would then need to pass
        #    quite a lot of stuff about the current swarms to the swarm class, *and* the [r] swarm
        #    needs updating too - lots of the logic is intertwined

        number_of_new_samples = 0

        limit = min(self.max_evol, len(self.swarms))

        # first identify those swarms who are at capacity, and therefore may be
        # considered for splitting off a member
        candidates = random.sample(self.swarms, limit)

        candidates = {x for x in candidates if x.number_of_particles >= self.swarm_size}

        if candidates:
            # select swarm at random
            [swarm] = random.sample(candidates, 1)

            # select an active swarm member at random (i.e. particle)
            k = random.randrange(swarm.swarm_size)

            particle_location = np.copy(swarm.history_locations[k, :])
            particle_value = swarm.history_values[k]

            # only look at splitting off member who is greater than tol_value
            # distance away -- otherwise will be merged right in again at the
            # next iteration
            if Nmmso._distance_to(particle_location, swarm.mode_location) > self.tol_val:

                mid_loc = 0.5 * (swarm.mode_location - particle_location) + particle_location

                mid_loc_val = self.problem.fitness(mid_loc)
                if self.listener is not None:
                    self.listener.location_evaluated(mid_loc, mid_loc_val)

                # if valley between, then hive off the old swarm member to create new swarm
                if mid_loc_val < particle_value:

                    new_swarm = self._new_swarm()

                    # allocate new
                    new_swarm.mode_location = particle_location
                    new_swarm.mode_value = particle_value

                    new_swarm.history_locations[0, :] = particle_location
                    new_swarm.history_values[0] = particle_value

                    new_swarm.pbest_locations[0, :] = particle_location
                    new_swarm.pbest_values[0] = particle_value

                    new_swarm.changed = True
                    new_swarm.converged = False

                    self._add_swarm(new_swarm)

                    # remove from existing swarm and replace with mid_eval
                    swarm.history_locations[k, :] = mid_loc
                    swarm.history_values[k] = mid_loc_val
                    swarm.pbest_locations[k, :] = mid_loc
                    swarm.pbest_values[k] = mid_loc_val

                    temp_vel = self.min - 1
                    d = Nmmso._distance_to(swarm.mode_location, particle_location)
                    reject = 0
                    while np.any(temp_vel < self.min) or np.any(temp_vel > self.max):
                        temp_vel = swarm.mode_location + \
                                   Nmmso.uniform_sphere_points(1, self.num_dimensions)[0] * (d / 2)

                        reject += 1
                        if reject > 20:
                            temp_vel = \
                                np.random.random(self.num_dimensions) * (self.max - self.min) + \
                                self.min

                    swarm.velocities[k, :] = temp_vel

                    if self.listener is not None:
                        self.listener.hiving_new_swarm(new_swarm, swarm)

                elif mid_loc_val > swarm.mode_value:
                    swarm.update_location_and_value(mid_loc, mid_loc_val)

                number_of_new_samples = 1

        return number_of_new_samples

    def _random_new(self):
        number_rand_modes = 1

        x = np.random.random_sample(np.shape(self.max)) * (self.max - self.min) + self.min

        new_swarm = self._new_swarm()
        new_swarm.changed = True
        new_swarm.converged = False
        new_swarm.new_location = x
        new_swarm.evaluate_first()
        self._add_swarm(new_swarm)

        if self.listener is not None:
            self.listener.swarm_created_at_random(new_swarm)

        return number_rand_modes

    @staticmethod
    def uniform_sphere_points(n, d):
        """
        Constructs uniform sphere points.

        Internal algorithm unity function.  Not to be called by users.

        Parameters:
        -----------

        n : int
            The number of points to construct

        d : int
            The number of dimensions of each point.

        Returns:
        numpy array
            2D array of uniform sphere points.

        """

        import operator

        z = np.random.normal(size=(n, d))

        # keepdims arg seems important; matlab explicitly seems to do this
        r1 = np.sqrt(np.sum(pow(z, 2), 1, keepdims=True))
        reps = np.tile(r1, (1, d))
        # using the function to make clear that it's element wise division
        # (/ symbol sometimes used for mrdivide)
        x = np.divide(z, reps)
        r = pow(np.random.rand(n, 1), (operator.truediv(1, d)))
        res = np.multiply(x, np.tile(r, (1, d)))  # np.multiply is element-wise
        return res

    def _add_swarm(self, swarm):
        self.swarms.add(swarm)
        self.next_swarm_id += 1

    def _new_swarm(self):
        return s.Swarm(self.next_swarm_id, self.swarm_size, self.problem, self.listener)



from multiprocessing import Process, Queue


def multiprocessor_process(problem, task_queue, result_queue):
    """
    Worker task main function.

    Arguments
    ---------

    problem
        problem object with a fitness method to invoke.
    task_queue : Queue
        Queue on which to receive tasks.
    result_queue : Queue
        Queue on which to return results.

    """
    while True:
        task = task_queue.get()
        task_id = task[0]
        task_loc = task[1]
        if task_id == -1:
            break

        # do some work
        res = problem.fitness(task_loc)
        result_queue.put([task_id, res])


class MultiprocessorFitnessCaller:
    """
    Fitness caller used for multiprocessor parallelism.

    Arguments
    ---------

    num_workers : int
        Number of worker nodes to create.

    """
    def __init__(self, num_workers):
        self.num_workers = num_workers
        self.problem = None

        self.total_tasks = 0
        self.total_groups = 0
        self.max_group_size = 0

        self.tasks = []

        self.num_calls = 0
        self.max_nodes = 0
        self.min_num_calls = 0

        self.num_workers = self.num_workers
        self.processes = []

        self.task_queue = Queue()
        self.result_queue = Queue()

    def set_problem(self, problem):
        """
        Sets the problem object to use to calculate the fitness.

        Arguments
        ---------

        problem
            Problem object implementing the fitness method.
        """
        for _ in range(self.num_workers):
            p = Process(target=multiprocessor_process,
                        args=(problem, self.task_queue, self.result_queue))
            p.start()
            self.processes.append(p)

    def add(self, location, userdata):
        """
        Add a location to be evaluated.

        Arguments
        ---------

        location : numpy array
            Location to be evaluated.

        userdata
            Userdata to be returned with the evaluation result.
        """
        self.tasks.append([location, userdata])

    def evaluate(self):
        """
        Evaluates all the locations.

        Returns
        -------

        list of (location, value, userdate) tuples
            Tuples containing the location, value and corresponding user data
        """
        num_tasks = len(self.tasks)

        self.total_tasks += num_tasks
        self.total_groups += 1
        if num_tasks > self.max_group_size:
            self.max_group_size = num_tasks

        for i in range(num_tasks):
            self.task_queue.put([i, self.tasks[i][0]]) # [index, loc]

        y = 0
        num_results = 0
        results = []
        while num_results < num_tasks:
            result = self.result_queue.get()  # [ index, y]
            index = result[0]
            y = result[1]
            results.append((self.tasks[index][0], y, self.tasks[index][1],))
            num_results += 1

        self.tasks = []
        return results

    def finish(self):
        """
        Terminates the fitness caller.
        """
        for p in self.processes:
            self.task_queue.put([-1, -1])
        for p in self.processes:
            p.join()



import random
import numpy as np
import pynmmso as nmmso


class Swarm:
    """
    Represents a swarm in the NMMSO algorithm.

    Arguments
    ---------

    id : int
        Id used to refer to the swarm
    swarm_size : int
        Maximum number of particles in the swarm
    problem :
        Instance of the problem class. Must implement get_bounds and fitness functions.
    listener : subclass of nmmso.listeners.BaseListener
            Listener object to receive notification of events. Optional.


    Attributes
    ----------

    id : int
        A unique identification number of this swarm.
    mode_location : numpy array
        The location of this mode.
    mode_value : float
        The fitness of the mode location.
    number_of_particles : int
        Number of particles in the swarm.
    history_locations : 2D Numpy array
        The current locations of each particle in the swarm.
    history_values : 1D Numpy array
        The fitness values for current locations of each particle in the swarm.
    velocities : 2D Numpy array
        Current velocity of each particle in the swarm.
    pbest_location : 2D Numpy array
        The best location discovered for each particle.
    pbest_value : 1D Numpy array
        The fitness value associated with the best location for each particle in the swarm.

    """
    def __init__(self, id, swarm_size, problem, listener=None):
        self.id = id
        self.swarm_size = swarm_size
        self.problem = problem
        self.listener = listener

        self.mn = np.array(problem.get_bounds()[0])
        self.mx = np.array(problem.get_bounds()[1])

        self.changed = True
        self.converged = False
        self.num_dimensions = len(self.mn)

        self.mode_location = None  # Will be populated later on
        self.new_location = None   # Will be populated later on
        self.mode_value = None     # Will be populated later on

        # Initialize locations for swarm elements

        # current locations of swarm
        self.history_locations = np.zeros((self.swarm_size, self.num_dimensions))
        # current values of swarm
        self.history_values = np.full(self.swarm_size, -np.inf)

        # current best locations of swarm
        self.pbest_locations = np.zeros((self.swarm_size, self.num_dimensions))
        # current best values of swarm
        self.pbest_values = np.full(self.swarm_size, -np.inf)

        self.velocities = np.zeros((swarm_size, self.num_dimensions))
        self.number_of_particles = 1

        self.shifted_loc = None  # Will be populated later on
        self.dist = None  # Will be populated later on

    def set_initial_location(self):
        """Sets the initial location of a swarm."""
        self.changed = True
        self.new_location = (np.random.rand(self.num_dimensions) * (self.mx-self.mn)) + self.mn
        # random initial velocities of swarm
        self.velocities[0, :] = (np.random.rand(self.num_dimensions) * (self.mx-self.mn)) + self.mn

    def set_arbitrary_distance(self):
        """Set an arbitrary distance - this is done when we only have one swarm"""
        self.dist = np.min(self.mx-self.mn)

    def increment(self):
        """ Increments the swarm. """
        new_location = self.mn - 1

        d = self.dist
        shifted = False
        omega = 0.1
        reject = 0

        r = random.randrange(self.swarm_size)   # select particle at random to move

        while np.sum(new_location < self.mn) > 0 or np.sum(new_location > self.mx) > 0:

            # if swarm is not yet at capacity, simply add a new particle
            if self.number_of_particles < self.swarm_size:
                usp = nmmso.Nmmso.uniform_sphere_points(1, self.num_dimensions)[0]
                new_location = self.mode_location + usp * (d/2)
            else:
                # move an existing particle
                shifted = True
                self.shifted_loc = r
                r1 = np.random.rand(self.num_dimensions)
                r2 = np.random.rand(self.num_dimensions)

                temp_vel = omega * self.velocities[self.shifted_loc, :] + \
                           2.0 * r1 * \
                           (self.mode_location - self.history_locations[self.shifted_loc, :]) + \
                           2.0 * r2 * \
                           (self.pbest_locations[self.shifted_loc, :] -
                            self.history_locations[self.shifted_loc, :])

                if reject > 20:
                    # if we keep rejecting then put at extreme any violating design parameters
                    i_max = np.flatnonzero(
                        np.asarray(
                            self.history_locations[self.shifted_loc, :] + temp_vel > self.mx))
                    i_min = np.flatnonzero(
                        np.asarray(
                            self.history_locations[self.shifted_loc, :] + temp_vel < self.mn))
                    if i_max.size > 0:
                        temp_vel[i_max] = \
                            np.random.rand(i_max.size) * \
                            (self.mx[i_max] - self.history_locations[self.shifted_loc, i_max])
                    if i_min.size > 0:
                        temp_vel[i_min] = \
                            np.random.rand(i_min.size) * \
                            (self.history_locations[self.shifted_loc, i_min] - self.mn[i_min])

                new_location = self.history_locations[self.shifted_loc, :] + temp_vel
                reject = reject + 1

        if shifted:
            self.velocities[self.shifted_loc, :] = temp_vel
        else:
            # otherwise initialise velocity in sphere based on distance from gbest to next
            # closest mode
            self.number_of_particles = self.number_of_particles + 1
            self.shifted_loc = self.number_of_particles - 1
            temp_vel = self.mn - 1

            reject = 0
            while np.sum(temp_vel < self.mn) > 0 or np.sum(temp_vel > self.mx) > 0:
                temp_vel = \
                    self.mode_location + \
                    nmmso.Nmmso.uniform_sphere_points(1, self.num_dimensions)[0] * (d / 2)
                reject = reject + 1
                if reject > 20:  # resolve if keep rejecting
                    temp_vel = np.random.rand(self.num_dimensions)*(self.mx-self.mn) + self.mn

            self.velocities[self.shifted_loc, :] = temp_vel

        self.new_location = new_location

        if self.listener is not None:
            if shifted:
                self.listener.swarm_moved_particle(self)
            else:
                self.listener.swarm_added_particle(self)

    def initialise_with_uniform_crossover(self, swarm1, swarm2):
        """
        Initialise a new swarm with the uniform crossover of the given swarms.

        Arguments
        ---------

        swarm1 : Swarm
        swarm2 : Swarm
        """
        self.new_location, _ = Swarm.uni(swarm1.mode_location, swarm2.mode_location)
        self.evaluate_first()
        self.changed = True
        self.converged = False

    def distance_to(self, swarm):
        """
        Euclidean distance between this swarm and the given swarm, based on their mode locations.

        Returns
        -------
        float
            The distance between the two swarms.
        """
        return np.linalg.norm(self.mode_location-swarm.mode_location)

    def merge(self, swarm):
        """
        Merges the give swarm into this swarm.

        Arguments
        ----------

        swarm : Swarm
            Swarm to merge into this swarm.

        """
        n1 = self.number_of_particles
        n2 = swarm.number_of_particles

        if n1 + n2 < self.swarm_size:
            # simplest solution, where the combined active members of both populations
            # are below the total size they can grow to
            self.number_of_particles = n1 + n2
            self.history_locations[n1:n1 + n2, :] = swarm.history_locations[0:n2, :]
            self.history_values[n1:n1 + n2] = swarm.history_values[0:n2]
            self.pbest_locations[n1:n1 + n2, :] = swarm.pbest_locations[0:n2, :]
            self.pbest_values[n1:n1 + n2] = swarm.pbest_values[0:n2]
            self.velocities[n1:n1 + n2, :] = swarm.velocities[0:n2, :]
        else:
            # select best out of combines population, based on current location (rather than pbest)
            self.number_of_particles = self.swarm_size
            temp_h_loc = \
                np.concatenate((self.history_locations[0:n1, :], swarm.history_locations[0:n2, :]))
            temp_h_v = \
                np.concatenate((self.history_values[0:n1], swarm.history_values[0:n2]))
            temp_p_loc = \
                np.concatenate((self.pbest_locations[0:n1, :], swarm.pbest_locations[0:n2, :]))
            temp_p_v = np.concatenate((self.pbest_values[0:n1], swarm.pbest_values[0:n2]))
            temp_vel = np.concatenate((self.velocities[0:n1, :], swarm.velocities[0:n2, :]))


            # get the indices of highest values
            I = np.argsort(temp_h_v)[len(temp_h_v) - self.swarm_size:]

            self.history_locations = temp_h_loc[I, :]
            self.history_values = temp_h_v[I]
            self.pbest_locations = temp_p_loc[I, :]
            self.pbest_values = temp_p_v[I]
            self.velocities = temp_vel[I, :]

    def initialise_new_swarm_velocities(self):
        """Initialises velocities of a new swarm."""

        reject = 0
        temp_vel = self.mn - 1
        while np.sum(temp_vel < self.mn) > 0 or np.sum(temp_vel > self.mx) > 0:
            temp_vel = self.mode_location + \
                       nmmso.Nmmso.uniform_sphere_points(1, self.num_dimensions)[0] * \
                       (self.dist / 2)
            reject += 1

            if reject > 20:
                temp_vel = (np.random.rand(self.num_dimensions) * (self.mx-self.mn)) + self.mn

        self.velocities[0, :] = temp_vel

    def update_location_and_value(self, location, value):
        """
        Updates the location and value of this swarm.

        Arguments
        ---------

        location : numpy arrya
            New location of swarm
        value : float
            New fitness value at swarm location.
        """
        previous_location = self.mode_location
        previous_value = self.mode_value
        self.mode_location = location
        self.mode_value = value
        if self.listener is not None:
            self.listener.swarm_peak_changed(self, previous_location, previous_value)

    def evaluate_first(self):
        """
        Evaluates the new location.  This is the first evaluation so no need to examine
        if a shift has occurred
        """

        # new location is the only solution thus far in mode, so by definition
        # is also the mode estimate, and the only history thus far

        y = self.problem.fitness(self.new_location)

        if self.listener is not None:
            self.listener.location_evaluated(self.new_location, y)

        self.mode_location = self.new_location  # gbest location
        self.mode_value = y  # gbest value

        self.history_locations[0, :] = self.mode_location
        self.history_values[0] = y

        self.pbest_locations[0, :] = self.mode_location
        self.pbest_values[0] = y

    def evaluate(self, y):
        """
        Takes the value at the new location and updates the swarm statistics and history.

        Arguments
        ---------

        y : float
            fitness value at the new location.
        """
        if y > self.mode_value:
            self.update_location_and_value(self.new_location, y)
            self.changed = True

        self.history_locations[self.shifted_loc, :] = self.new_location
        self.history_values[self.shifted_loc] = y

        if y > self.pbest_values[self.shifted_loc]:
            self.pbest_values[self.shifted_loc] = y
            self.pbest_locations[self.shifted_loc, :] = self.new_location

    def find_nearest(self, swarms):
        """
        Finds the nearest swarm from the given set of swarms.

        Returns
        -------
        swarm
            The nearest swarm this this swarm.
        """
        best_swarm = None
        distance = np.inf

        for s in swarms:
            if self != s:
                d = np.sum((self.mode_location - s.mode_location) ** 2)
                if d < distance:
                    distance = d
                    best_swarm = s

        self.dist = np.sqrt(distance)  # track Euc distance to nearest neighbour

        return best_swarm, self.dist

    @staticmethod
    def uni(x1, x2):
        """
        Uniform binary crossover.

        Arguments
        ---------

        x1 : numpy array of parameters
        x2 : numpy array of parameters

        Returns:
        numpy array
            New array of parameters formed from uniform crossover.
        """

        # simulated binary crossover
        x_c = x1.copy()
        x_d = x2.copy()
        l = len(x1)
        r = np.flatnonzero(np.random.rand(l, 1) > 0.5)
        # ensure at least one is swapped
        if r.size == 0 or r.size == l:
            r = np.random.randint(l)

        x_c[r] = x2[r]
        x_d[r] = x1[r]

        return x_c, x_d



import numpy as np


class UniformRangeProblem:
    """
    Class used to wrap a problem class so that all parameters have a uniform range.

    Arguments
    ---------

    problem
        The problem object to be wrapped.
    """

    def __init__(self, problem):
        self.problem = problem
        mn, mx = problem.get_bounds()
        self.min = np.array(mn)
        mx = np.array(mx)
        self.range = mx - self.min

    def fitness(self, location):
        """
        Calls the fitness function.

        Arguments
        ---------

        location : numpy array
            The location to evaluate the fitness of.

        Returns
        -------

        float
            The fitness value at the given location.
        """
        return self.problem.fitness(self.remap_parameters(location))

    def get_bounds(self):
        """
        Gets the bounds of the problem.

        Returns
        -------
        numpy array, numpy array
            One array with the lower bounds and an array with the upper bounds.
        """
        return [0] * len(self.min), [1] * len(self.min)

    def remap_parameters(self, location):
        """
        Maps a given location in the new parameter space to a location in the
        parameter space of the wrapped problem.

        Arguments
        ---------

        location : numpy array
            The location in the uniform parameter space.

        Returns
        -------

        numpy array
            The corresponding location in the parameter space of the
            wrapped problem.

        """
        return (location * self.range) + self.min



from pynmmso.listeners.base_listener import BaseListener


class MultiListener(BaseListener):
    """
    Listener that contains multiple listeners and calls them for each event.
    """

    def __init__(self):
        super().__init__()
        self.listeners = []

    def add_listener(self, listener):
        """
        Adds a new listener.

        Arguments
        ---------

        listener
            The listener to add.
        """
        self.listeners.append(listener)

    def set_nmmso(self, nmmso):
        for listener in self.listeners:
            listener.set_nmmso(nmmso)

    def iteration_started(self):
        for listener in self.listeners:
            listener.iteration_started()

    def location_evaluated(self, location, value):
        for listener in self.listeners:
            listener.location_evaluated(location, value)

    def swarm_peak_changed(self, swarm, old_location, old_value):
        for listener in self.listeners:
            listener.swarm_peak_changed(swarm, old_location, old_value)

    def swarm_created_at_random(self, new_swarm):
        for listener in self.listeners:
            listener.swarm_created_at_random(new_swarm)

    def swarm_created_from_crossover(self, new_swarm, parent_swarm1, parent_swarm2):
        for listener in self.listeners:
            listener.swarm_created_from_crossover(new_swarm, parent_swarm1, parent_swarm2)

    def merging_started(self):
        for listener in self.listeners:
            listener.merging_started()

    def merged_close_swarms(self, swarm1, swarm2):
        for listener in self.listeners:
            listener.merged_close_swarms(swarm1, swarm2)

    def merged_saddle_swarms(self, swarm1, swarm2):
        for listener in self.listeners:
            listener.merged_saddle_swarms(swarm1, swarm2)

    def merging_ended(self):
        for listener in self.listeners:
            listener.merging_ended()

    def incrementing_swarms_started(self):
        for listener in self.listeners:
            listener.incrementing_swarms_started()

    def swarm_added_particle(self, swarm):
        for listener in self.listeners:
            listener.swarm_added_particle(swarm)

    def swarm_moved_particle(self, swarm):
        for listener in self.listeners:
            listener.swarm_moved_particle(swarm)

    def incrementing_swarms_ended(self):
        for listener in self.listeners:
            listener.incrementing_swarms_ended()

    def hiving_swams_started(self):
        for listener in self.listeners:
            listener.hiving_swams_started()

    def hiving_new_swarm(self, new_swarm, parent_swarm):
        for listener in self.listeners:
            listener.hiving_new_swarm(new_swarm, parent_swarm)

    def hiving_swarms_ended(self):
        for listener in self.listeners:
            listener.hiving_swarms_ended()

    def iteration_ended(
            self, n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples):
        for listener in self.listeners:
            listener.iteration_ended(
                n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples)

    def max_evaluations_reached(self):
        for listener in self.listeners:
            listener.max_evaluations_reached()



from pynmmso.listeners.base_listener import BaseListener


class TraceListener(BaseListener):
    """
    Listener used to trace the progress of the Nmmso algorithm.

    Arguments
    ---------

    level : int
        The amount of detail to output in the trace. Level 1 is the least information and
        level 5 is the most.  Default is level 2.

    """
    def __init__(self, level=2):
        self.nmmso = None
        self.iteration_number = 1
        self.evaluations = 0
        self.level = level
        super().__init__()

    def set_nmmso(self, nmmso):
        self.nmmso = nmmso

    def iteration_started(self):
        if self.level >= 3:
            print(80*"=")
            print("Starting iteration {}".format(self.iteration_number))

    def location_evaluated(self, location, value):
        self.evaluations += 1
        if self.level >= 5:
            print("Evaluation {}: location {}, value is {}".format(
                self.evaluations, location, value))

    def swarm_peak_changed(self, swarm, old_location, old_value):
        if self.level >= 3:
            print("Swarm {} has found a new peak at location {} with value {}, "
                  "old location was {} old value was {}".format(
                      swarm.id, swarm.mode_location, swarm.mode_value, old_location, old_value))

    def swarm_created_at_random(self, new_swarm):
        if self.level >= 3:
            print("Created swarm {} at random location {}, value is {}".format(
                new_swarm.id, new_swarm.mode_location, new_swarm.mode_value))

    def swarm_created_from_crossover(self, new_swarm, parent_swarm1, parent_swarm2):
        if self.level >= 3:
            print("Created swarm {} by crossover of swarms {} and {} at location {}, "
                  "value is {}".format(
                      new_swarm.id, parent_swarm1.id, parent_swarm2.id,
                      new_swarm.mode_location, new_swarm.mode_value, ))

    def merging_started(self):
        if self.level >= 4:
            print("Merging swarms...")

    def merged_close_swarms(self, swarm1, swarm2):
        if self.level >= 3:
            print("Merged swarm {} into swarm {} as they were close".format(swarm2.id, swarm1.id))

    def merged_saddle_swarms(self, swarm1, swarm2):
        if self.level >= 3:
            print("Merged swarm {} into swarm {} as midpoint was fitter".format(
                swarm2.id, swarm1.id))

    def merging_ended(self):
        if self.level >= 4:
            print("Finished merging swarms")

    def incrementing_swarms_started(self):
        if self.level >= 4:
            print("Incrementing swarms...")

    def swarm_added_particle(self, swarm):
        if self.level >= 4:
            print("Added particle to swarm {}, it now has {} particles".format(
                swarm.id, swarm.number_of_particles))

    def swarm_moved_particle(self, swarm):
        if self.level >= 4:
            print("Moved particle of swarm {}".format(swarm.id))

    def incrementing_swarms_ended(self):
        if self.level >= 4:
            print("Finished incrementing swarms")

    def hiving_swams_started(self):
        if self.level >= 4:
            print("Hiving swarms...")

    def hiving_new_swarm(self, new_swarm, parent_swarm):
        if self.level >= 3:
            print("Hiving new swarm {} from swarm {}".format(new_swarm.id, parent_swarm.id))

    def hiving_swarms_ended(self):
        if self.level >= 4:
            print("Finishing hiving swarms")

    def iteration_ended(
            self, n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples):
        total_this_iteration = \
            n_new_locations + n_mid_evals + n_evol_modes + n_rand_modes + n_hive_samples

        if self.level >= 1:
            print("Finished iteration {}, evaluations this iteration: {}, total evaluations: {}, "
                  "number of swarms: {}".format(
                      self.iteration_number,
                      total_this_iteration,
                      self.nmmso.evaluations,
                      len(self.nmmso.swarms)))

        if self.level >= 3:
            print("  This iteration: new location evals = {} mid evals = {} evol modes = {} "
                  "rand modes = {} hive samples = {}".format(
                      n_new_locations, n_mid_evals, n_evol_modes, n_rand_modes, n_hive_samples))

        if self.level >= 2:
            for swarm in self.nmmso.swarms:
                print("Swarm {} : location: {}  value {}".format(
                    swarm.id, swarm.mode_location, swarm.mode_value))

        self.iteration_number += 1

    def max_evaluations_reached(self):
        if self.level >= 1:
            print("Maximum number of evaluations reached.  Total evaluations: {}".format(
                self.nmmso.evaluations))



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()
        
setup(name='pynmmso',
      version='0.1',
      description='Multi-Modal Search with the Niching Migratory Multi-Swarm Optimizer',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering',
        'Operating System :: OS Independent',
      ],
      keywords='optimization, optimisation, optimizer, optimiser, multimodal, multi-modal, GA, genetic',
      url='https://github.com/EPCCed/NMMSO/wiki/NMMSO',
      author='Ally Hume',
      author_email='a.hume@epcc.ed.ac.uk',
      license='MIT',
      packages=['pynmmso','pynmmso.wrappers','pynmmso.listeners'],
      install_requires=[
          'numpy',
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      zip_safe=False)
